import click
from rich.console import Console
from rich.table import Column, Table

from cli.client import init_client
from cli.helpers.nucleus_url import nucleus_url
from cli.helpers.web_helper import launch_web_or_invoke


@click.group("datasets", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def datasets(ctx, web):
    """Datasets are the base collections of items in Nucleus

    https://dashboard.scale.com/nucleus/datasets
    """
    launch_web_or_invoke(
        sub_url="datasets", ctx=ctx, launch_browser=web, command=list_datasets
    )


@datasets.command("list")
@click.option(
    "-m", "--machine-readable", is_flag=True, help="Removes pretty printing"
)
def list_datasets(machine_readable):
    """List all available Datasets"""
    console = Console()
    with console.status("Finding your Datasets!", spinner="dots4"):
        client = init_client()
        all_datasets = client.datasets
        if machine_readable:
            table_params = {"box": None, "pad_edge": False}
        else:
            table_params = {
                "title": ":fire: Datasets",
                "title_justify": "left",
            }

        table = Table(
            "id", "Name", Column("url", overflow="fold"), **table_params
        )
        for ds in all_datasets:
            table.add_row(ds.id, ds.name, nucleus_url(ds.id))
    console.print(table)


@datasets.command("delete")
@click.option("--id", prompt=True)
@click.option(
    "--no-confirm-deletion",
    is_flag=True,
    help="WARNING: No confirmation for deletion",
)
@click.pass_context
def delete_dataset(ctx, id, no_confirm_deletion):
    """Delete a Dataset"""
    console = Console()
    client = init_client()
    id = id.strip()
    dataset = client.get_dataset(id)
    delete_string = ""
    if not no_confirm_deletion:
        delete_string = click.prompt(
            click.style(
                f"Type 'DELETE' to delete dataset: {dataset}", fg="red"
            )
        )
    if no_confirm_deletion or delete_string == "DELETE":
        client.delete_dataset(dataset.id)
        console.print(f":fire: :anguished: Deleted {id}")
    else:
        console.print(
            f":rotating_light: Refusing to delete {id}. Received '{delete_string}' instead of 'DELETE'"
        )
        ctx.abort()



import abc
import logging
from enum import Enum
from typing import List, Optional, Union

import numpy as np

from nucleus.annotation import AnnotationList, SegmentationAnnotation
from nucleus.metrics.base import MetricResult
from nucleus.metrics.filtering import (
    ListOfAndFilters,
    ListOfOrAndFilters,
    apply_filters,
)
from nucleus.metrics.segmentation_utils import (
    instance_mask_to_polys,
    rasterize_polygons_to_segmentation_mask,
    setup_iou_thresholds,
    transform_poly_codes_to_poly_preds,
)
from nucleus.prediction import PredictionList

from .base import Metric, ScalarResult
from .polygon_metrics import (
    PolygonAveragePrecision,
    PolygonIOU,
    PolygonPrecision,
    PolygonRecall,
)
from .segmentation_loader import (
    DummyLoader,
    InMemoryLoader,
    SegmentationMaskLoader,
)
from .segmentation_metrics import (
    SegmentationIOU,
    SegmentationMAP,
    SegmentationPrecision,
    SegmentationRecall,
)


class SegToPolyMode(str, Enum):
    GENERATE_GT_FROM_POLY = "gt_from_poly"
    GENERATE_PRED_POLYS_FROM_MASK = "gt_from_poly"


class SegmentationMaskToPolyMetric(Metric):
    def __init__(
        self,
        enforce_label_match: bool = False,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes PolygonMetric abstract object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Default False
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        # Since segmentation annotations are very different from everything else we can't rely on the upper filtering
        super().__init__(None, None)
        self._annotation_filters = annotation_filters
        self._prediction_filters = prediction_filters
        self.enforce_label_match = enforce_label_match
        assert 0 <= confidence_threshold <= 1
        self.confidence_threshold = confidence_threshold
        self.loader: SegmentationMaskLoader = DummyLoader()
        self.mode = mode

    def call_metric(
        self, annotations: AnnotationList, predictions: PredictionList
    ) -> MetricResult:
        assert (
            len(predictions.segmentation_predictions) <= 1
        ), f"Expected only one segmentation mask, got {predictions.segmentation_predictions}"
        prediction = (
            predictions.segmentation_predictions[0]
            if predictions.segmentation_predictions
            else None
        )
        annotations.polygon_annotations = apply_filters(
            annotations.polygon_annotations, self._annotation_filters  # type: ignore
        )
        annotations.box_annotations = apply_filters(
            annotations.box_annotations, self._annotation_filters  # type: ignore
        )
        predictions.segmentation_predictions = apply_filters(
            predictions.segmentation_predictions, self._prediction_filters  # type: ignore
        )
        if prediction:
            if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
                pred_img = self.loader.fetch(prediction.mask_url)
                ann_img, segments = rasterize_polygons_to_segmentation_mask(
                    annotations.polygon_annotations
                    + annotations.box_annotations,  # type:ignore
                    pred_img.shape,
                )
                # TODO: apply Segmentation filters after?
                annotations.segmentation_annotations = [
                    SegmentationAnnotation(
                        "__no_url",
                        annotations=segments,
                        reference_id=annotations.polygon_annotations[
                            0
                        ].reference_id,
                    )
                ]
                return self.call_segmentation_metric(
                    annotations,
                    np.asarray(ann_img),
                    predictions,
                    np.asarray(pred_img),
                )
            elif self.mode == SegToPolyMode.GENERATE_PRED_POLYS_FROM_MASK:
                pred_img = self.loader.fetch(prediction.mask_url)
                pred_value, pred_polys = instance_mask_to_polys(
                    np.asarray(pred_img)
                )  # typing: ignore
                code_to_label = {
                    s.index: s.label for s in prediction.annotations
                }
                poly_predictions = transform_poly_codes_to_poly_preds(
                    prediction.reference_id,
                    pred_value,
                    pred_polys,
                    code_to_label,
                )
                return self.call_poly_metric(
                    annotations,
                    PredictionList(polygon_predictions=poly_predictions),
                )
            else:
                raise RuntimeError(
                    f"Misonconfigured class. Got mode '{self.mode}', expected one of {list(SegToPolyMode)}"
                )
        else:
            return ScalarResult(0, weight=0)

    def call_segmentation_metric(
        self,
        annotations: AnnotationList,
        ann_img: "np.ndarray",
        predictions: PredictionList,
        pred_img: "np.ndarray",
    ):
        metric = self.configure_metric()
        metric.loader = InMemoryLoader(
            {
                annotations.segmentation_annotations[0].mask_url: ann_img,
                predictions.segmentation_predictions[0].mask_url: pred_img,
            }
        )
        return metric(annotations, predictions)

    def call_poly_metric(
        self, annotations: AnnotationList, predictions: PredictionList
    ):
        metric = self.configure_metric()
        return metric(annotations, predictions)

    def aggregate_score(self, results: List[MetricResult]) -> ScalarResult:
        metric = self.configure_metric()
        return metric.aggregate_score(results)  # type: ignore

    @abc.abstractmethod
    def configure_metric(self):
        pass


class SegmentationToPolyIOU(SegmentationMaskToPolyMetric):
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.0,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes PolygonIOU object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.0
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters,
            prediction_filters,
            mode,
        )

    def configure_metric(self):
        if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
            metric = SegmentationIOU(
                self.annotation_filters,
                self.prediction_filters,
                self.iou_threshold,
            )
        else:
            metric = PolygonIOU(
                self.enforce_label_match,
                self.iou_threshold,
                self.confidence_threshold,
                self.annotation_filters,
                self.prediction_filters,
            )
        return metric


class SegmentationToPolyPrecision(SegmentationMaskToPolyMetric):
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes SegmentationToPolyPrecision object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters,
            prediction_filters,
            mode,
        )

    def configure_metric(self):
        if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
            metric = SegmentationPrecision(
                self.annotation_filters,
                self.prediction_filters,
                self.iou_threshold,
            )
        else:
            metric = PolygonPrecision(
                self.enforce_label_match,
                self.iou_threshold,
                self.confidence_threshold,
                self.annotation_filters,
                self.prediction_filters,
            )
        return metric


class SegmentationToPolyRecall(SegmentationMaskToPolyMetric):
    """Calculates the recall between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonRecall

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonRecall()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        enforce_label_match: bool = False,
        iou_threshold: float = 0.5,
        confidence_threshold: float = 0.0,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes PolygonRecall object.

        Args:
            enforce_label_match: whether to enforce that annotation and prediction labels must match. Defaults to False
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            confidence_threshold: minimum confidence threshold for predictions. Must be in [0, 1]. Default 0.0
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        super().__init__(
            enforce_label_match,
            confidence_threshold,
            annotation_filters,
            prediction_filters,
            mode,
        )

    def configure_metric(self):
        if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
            metric = SegmentationRecall(
                self.annotation_filters,
                self.prediction_filters,
                self.iou_threshold,
            )
        else:
            metric = PolygonRecall(
                self.enforce_label_match,
                self.iou_threshold,
                self.confidence_threshold,
                self.annotation_filters,
                self.prediction_filters,
            )
        return metric


class SegmentationToPolyAveragePrecision(SegmentationMaskToPolyMetric):
    """Calculates the average precision between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonAveragePrecision

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonAveragePrecision(label="car")
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        label,
        iou_threshold: float = 0.5,
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes PolygonRecall object.

        Args:
            iou_threshold: IOU threshold to consider detection as valid. Must be in [0, 1]. Default 0.5
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        assert (
            0 <= iou_threshold <= 1
        ), "IoU threshold must be between 0 and 1."
        self.iou_threshold = iou_threshold
        self.label = label
        super().__init__(
            enforce_label_match=False,
            confidence_threshold=0,
            annotation_filters=annotation_filters,
            prediction_filters=prediction_filters,
        )

    def configure_metric(self):
        if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
            # TODO(gunnar): Add a label filter
            metric = SegmentationPrecision(
                self.annotation_filters,
                self.prediction_filters,
                self.iou_threshold,
            )
        else:
            metric = PolygonAveragePrecision(
                self.label,
                self.iou_threshold,
                self.annotation_filters,
                self.prediction_filters,
            )
        return metric


class SegmentationToPolyMAP(SegmentationMaskToPolyMetric):
    """Calculates the mean average precision between box or polygon annotations and predictions.
    ::

        from nucleus import BoxAnnotation, Point, PolygonPrediction
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import PolygonMAP

        box_anno = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        polygon_pred = PolygonPrediction(
            label="bus",
            vertices=[Point(100, 100), Point(150, 200), Point(200, 100)],
            reference_id="image_2",
            annotation_id="image_2_bus_polygon_1",
            confidence=0.8,
            metadata={"vehicle_color": "yellow"}
        )

        annotations = AnnotationList(box_annotations=[box_anno])
        predictions = PredictionList(polygon_predictions=[polygon_pred])
        metric = PolygonMAP()
        metric(annotations, predictions)
    """

    # TODO: Remove defaults once these are surfaced more cleanly to users.
    def __init__(
        self,
        iou_threshold: float = -1,
        iou_thresholds: Union[List[float], str] = "coco",
        annotation_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        prediction_filters: Optional[
            Union[ListOfOrAndFilters, ListOfAndFilters]
        ] = None,
        mode: SegToPolyMode = SegToPolyMode.GENERATE_GT_FROM_POLY,
    ):
        """Initializes PolygonRecall object.

        Args:
            iou_thresholds: IOU thresholds to check AP at
            annotation_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
            prediction_filters: Filter predicates. Allowed formats are:
                ListOfAndFilters where each Filter forms a chain of AND predicates.
                    or
                ListOfOrAndFilters where Filters are expressed in disjunctive normal form (DNF), like
                [[MetadataFilter("short_haired", "==", True), FieldFilter("label", "in", ["cat", "dog"]), ...].
                DNF allows arbitrary boolean logical combinations of single field predicates. The innermost structures
                each describe a single column predicate. The list of inner predicates is interpreted as a conjunction
                (AND), forming a more selective `and` multiple field predicate.
                Finally, the most outer list combines these filters as a disjunction (OR).
        """
        if iou_threshold:
            logging.warning(
                "Got deprecated parameter 'iou_threshold'. Ignoring it."
            )
        self.iou_thresholds = setup_iou_thresholds(iou_thresholds)
        super().__init__(
            False, 0, annotation_filters, prediction_filters, mode
        )

    def configure_metric(self):
        if self.mode == SegToPolyMode.GENERATE_GT_FROM_POLY:
            # TODO(gunnar): Add a label filter
            metric = SegmentationMAP(
                self.annotation_filters,
                self.prediction_filters,
                self.iou_thresholds,
            )
        else:

            def patched_average_precision(annotations, predictions):
                ap_per_threshold = []
                labels = [p.label for p in predictions.polygon_predictions]
                for threshold in self.iou_thresholds:
                    ap_per_label = []
                    for label in labels:
                        call_metric = PolygonAveragePrecision(
                            label,
                            iou_threshold=threshold,
                            annotation_filters=self.annotation_filters,
                            prediction_filters=self.prediction_filters,
                        )
                        result = call_metric(annotations, predictions)
                        ap_per_label.append(result.value)  # type: ignore
                    ap_per_threshold = np.mean(ap_per_label)

                thresholds = np.concatenate([[0], self.iou_thresholds, [1]])
                steps = np.diff(thresholds)
                mean_ap = (
                    np.array(ap_per_threshold + [ap_per_threshold[-1]]) * steps
                ).sum()
                return ScalarResult(mean_ap)

            metric = patched_average_precision
        return metric



import click
from rich.console import Console
from rich.pretty import pretty_repr
from rich.table import Column, Table

from cli.client import init_client
from cli.helpers.nucleus_url import nucleus_url
from cli.helpers.web_helper import launch_web_or_invoke


@click.group("models", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def models(ctx, web):
    """Models help you store and access your ML model data

    https://dashboard.scale.com/nucleus/models
    """
    launch_web_or_invoke("models", ctx, web, list_models)


STRING_REPLACEMENTS = {
    "\\n": "\n",
    "\\t": "\t",
    '\\"': '"',
}


def json_string_to_string(s: str) -> str:
    for key, val in STRING_REPLACEMENTS.items():
        s = s.replace(key, val)
    return s


@models.command("calculate-metrics")
def metrics():
    import questionary

    client = init_client()
    models = client.models
    prompt_to_id = {f"{m.id}: {m.name}": m.id for m in models}
    ans = questionary.select(
        "What model do you want to run metrics for?",
        choices=list(prompt_to_id.keys()),
    ).ask()
    model_id = prompt_to_id[ans]
    jobs = client.validate.metrics(model_id)
    console = Console()
    with console.status("Calculating metrics"):
        for job in jobs:
            job.sleep_until_complete(False)

    if len(job.errors()) == 0:
        status = job.status()
        click.echo(click.style("Done", fg="green"))
        console.print(pretty_repr(status))
    else:
        click.echo(
            click.style("Encountered errors during running", fg="green")
        )
        for error in job.errors():
            click.echo(json_string_to_string(error))


@models.command("list")
def list_models():
    """List your Models"""
    console = Console()
    with console.status("Finding your Models!", spinner="dots4"):
        client = init_client()
        table = Table(
            Column("id", overflow="fold", min_width=24),
            "name",
            Column("url", overflow="fold"),
        )
        models = client.models
        for m in models:
            table.add_row(m.id, m.name, nucleus_url(m.id))
    console.print(table)



from pydantic import BaseModel  # pylint: disable=no-name-in-module


class ImmutableModel(BaseModel):
    class Config:
        allow_mutation = False


class DictCompatibleModel(BaseModel):
    """Backwards compatible wrapper where we transform dictionaries into Pydantic Models

    Allows us to access model.key with model["key"].
    """

    def __getitem__(self, key):
        return getattr(self, key)


class DictCompatibleImmutableModel(ImmutableModel):
    """Backwards compatible wrapper where we transform dictionaries into Pydantic Models

    Allows us to access model.key with model["key"].
    """

    def __getitem__(self, key):
        return getattr(self, key)



import io
from typing import Any, Callable, Dict, List, Optional, Type

from PIL import Image, ImageDraw
from pydantic import BaseModel, Extra, ValidationError

# From scaleapi/server/src/lib/select/api/types.ts
# These classes specify how user models must pass output to Launch + Nucleus.


class PointModel(BaseModel, extra=Extra.forbid):
    x: float
    y: float


class BoxGeometryModel(BaseModel, extra=Extra.forbid):
    x: float
    y: float
    width: float
    height: float


class BoxAnnotationModel(BaseModel, extra=Extra.forbid):
    geometry: BoxGeometryModel
    type: str
    label: Optional[str] = None
    confidence: Optional[float] = None
    classPdf: Optional[Dict[str, float]] = None
    metadata: Optional[Dict[str, Any]] = None


class NoneGeometryModel(BaseModel, extra=Extra.forbid):
    pass


class CategoryAnnotationModel(BaseModel, extra=Extra.forbid):
    geometry: NoneGeometryModel
    type: str
    label: Optional[str] = None
    confidence: Optional[float] = None
    classPdf: Optional[Dict[str, float]] = None
    metadata: Optional[Dict[str, Any]] = None


class LineGeometryModel(BaseModel, extra=Extra.forbid):
    vertices: List[PointModel]


class LineAnnotationModel(BaseModel, extra=Extra.forbid):
    geometry: LineGeometryModel
    type: str
    label: Optional[str] = None
    confidence: Optional[float] = None
    classPdf: Optional[Dict[str, float]] = None
    metadata: Optional[Dict[str, Any]] = None


class PolygonGeometryModel(BaseModel, extra=Extra.forbid):
    vertices: List[PointModel]


class PolygonAnnotationModel(BaseModel, extra=Extra.forbid):
    geometry: PolygonGeometryModel
    type: str
    label: Optional[str] = None
    confidence: Optional[float] = None
    classPdf: Optional[Dict[str, float]] = None
    metadata: Optional[Dict[str, Any]] = None


def verify_output(
    annotation_list: List[Dict[str, Any]],
    model: Type[BaseModel],
    annotation_type: str,
):
    for annotation in annotation_list:
        try:
            model.parse_obj(annotation)
        except ValidationError as e:
            raise ValueError("Failed validation") from e
        if annotation["type"] != annotation_type:
            raise ValueError(
                f"Bounding box type {annotation['type']} should equal {annotation_type}"
            )


def verify_box_output(bbox_list):
    annotation_type = "box"
    return verify_output(
        bbox_list,
        BoxAnnotationModel,
        annotation_type,
    )


def verify_category_output(category_list):
    """I think the annotation needs to be a list with a single element in the Launch+Nucleus sfn."""
    annotation_type = "category"
    return verify_output(
        category_list, CategoryAnnotationModel, annotation_type
    )


def verify_line_output(line_list):
    annotation_type = "line"
    return verify_output(
        line_list,
        LineAnnotationModel,
        annotation_type,
    )


def verify_polygon_output(polygon_list):
    annotation_type = "polygon"
    return verify_output(
        polygon_list,
        PolygonAnnotationModel,
        annotation_type,
    )


def _run_model(
    input_bytes: bytes,
    load_predict_fn: Callable,
    load_model_fn: Optional[Callable],
    model: Optional[Any],
):
    if not (model is None) ^ (load_model_fn is None):
        raise ValueError(
            "Exactly one of `model` and `load_model_fn` must not be None."
        )

    if load_model_fn:
        model = load_model_fn()

    predict_fn = load_predict_fn(model)
    return predict_fn(input_bytes)


_FILL_COLOR = (0, 255, 0, 50)
_OUTLINE_COLOR = (0, 255, 0, 255)


def visualize_box_launch_bundle(
    img_file: str,
    load_predict_fn: Callable,
    load_model_fn: Callable = None,
    model: Any = None,
    show_image: bool = False,
    max_annotations: int = 5,
) -> Image:
    """
    Run this function locally to visualize what your Launch bundle will do on a local image
    Intended to verify that your Launch bundle returns annotations in the correct format, as well as sanity check
    any coordinate systems used for the image.
    Will display the image in a separate window if show_image == True.
    Returns the image as well.

    Parameters:
        img_file: The path to a local image file.
        load_predict_fn: The load_predict_fn as part of your Launch bundle
        load_model_fn: The load_model_fn as part of your Launch bundle
        model: The model as part of your Launch bundle. Note: exactly one of load_model_fn and model must be specified
        show_image: Whether to automatically pop up the image + predictions in a separate window. Can be useful in a
          script.
        max_annotations: How many annotations you want to draw

    Returns:
        Image: The image with annotations drawn on top.
    """
    # Basically do the same thing as what Launch does but locally

    with open(img_file, "rb") as f:
        img_bytes = f.read()

    output = _run_model(img_bytes, load_predict_fn, load_model_fn, model)
    verify_box_output(output)

    image = Image.open(io.BytesIO(img_bytes))
    draw = ImageDraw.Draw(image, "RGBA")
    for bbox in output[:max_annotations]:
        geo = bbox["geometry"]
        x, y, w, h = geo["x"], geo["y"], geo["width"], geo["height"]
        draw.rectangle(
            [(x, y), (x + w, y + h)], outline=_OUTLINE_COLOR, fill=_FILL_COLOR
        )

    if show_image:
        image.show()

    return image


def run_category_launch_bundle(
    img_file: str,
    load_predict_fn: Callable,
    load_model_fn: Callable = None,
    model: Any = None,
):
    """
    Run this function locally to test if your image categorization model returns a format consumable by Launch + Nucleus
    Parameters:
        img_file: The path to a local image file.
        load_predict_fn: The load_predict_fn as part of your Launch bundle
        load_model_fn: The load_model_fn as part of your Launch bundle
        model: The model as part of your Launch bundle. Note: exactly one of load_model_fn and model must be specified
    Returns:
        The raw output (as a json) of your categorization model.
    """
    with open(img_file, "rb") as f:
        img_bytes = f.read()

    output = _run_model(img_bytes, load_predict_fn, load_model_fn, model)
    verify_category_output(output)
    return output


def visualize_line_launch_bundle(
    img_file: str,
    load_predict_fn: Callable,
    load_model_fn: Callable = None,
    model: Any = None,
    show_image: bool = False,
    max_annotations: int = 5,
) -> Image:
    """
    Run this function locally to visualize what your Launch bundle will do on a local image
    Intended to verify that your Launch bundle returns annotations in the correct format, as well as sanity check
    any coordinate systems used for the image.
    Will display the image in a separate window if show_image == True.
    Returns the image as well.

    Parameters:
        img_file: The path to a local image file.
        load_predict_fn: The load_predict_fn as part of your Launch bundle
        load_model_fn: The load_model_fn as part of your Launch bundle
        model: The model as part of your Launch bundle. Note: exactly one of load_model_fn and model must be specified
        show_image: Whether to automatically pop up the image + predictions in a separate window. Can be useful in a
          script.
        max_annotations: How many annotations you want to draw

    Returns:
        Image: The image with annotations drawn on top.
    """
    # Basically do the same thing as what Launch does but locally

    with open(img_file, "rb") as f:
        img_bytes = f.read()

    output = _run_model(img_bytes, load_predict_fn, load_model_fn, model)
    verify_line_output(output)

    image = Image.open(io.BytesIO(img_bytes))
    draw = ImageDraw.Draw(image, "RGBA")
    for bbox in output[:max_annotations]:
        geo = bbox["geometry"]
        vertices = [(v["x"], v["y"]) for v in geo["vertices"]]
        draw.line(vertices, fill=_OUTLINE_COLOR)

    if show_image:
        image.show()

    return image


def visualize_polygon_launch_bundle(
    img_file: str,
    load_predict_fn: Callable,
    load_model_fn: Callable = None,
    model: Any = None,
    show_image: bool = False,
    max_annotations: int = 5,
) -> Image:
    """
    Run this function locally to visualize what your Launch bundle will do on a local image
    Intended to verify that your Launch bundle returns annotations in the correct format, as well as sanity check
    any coordinate systems used for the image.
    Will display the image in a separate window if show_image == True.
    Returns the image as well.

    Parameters:
        img_file: The path to a local image file.
        load_predict_fn: The load_predict_fn as part of your Launch bundle
        load_model_fn: The load_model_fn as part of your Launch bundle
        model: The model as part of your Launch bundle. Note: exactly one of load_model_fn and model must be specified
        show_image: Whether to automatically pop up the image + predictions in a separate window. Can be useful in a
          script.
        max_annotations: How many annotations you want to draw

    Returns:
        Image: The image with annotations drawn on top.
    """
    # Basically do the same thing as what Launch does but locally

    with open(img_file, "rb") as f:
        img_bytes = f.read()

    output = _run_model(img_bytes, load_predict_fn, load_model_fn, model)
    verify_polygon_output(output)

    image = Image.open(io.BytesIO(img_bytes))
    draw = ImageDraw.Draw(image, "RGBA")
    for bbox in output[:max_annotations]:
        geo = bbox["geometry"]
        vertices = [(v["x"], v["y"]) for v in geo["vertices"]]
        draw.polygon(vertices, outline=_OUTLINE_COLOR, fill=_FILL_COLOR)

    if show_image:
        image.show()

    return image



import abc
from typing import Any, Dict

from ..constants import ThresholdComparison
from ..data_transfer_objects.eval_function import (
    EvalFunctionEntry,
    EvaluationCriterion,
)


class EvalFunctionConfig(abc.ABC):
    """Abstract base class for concrete implementations of EvalFunctionsConfigs

    Operating on this class with comparison operators produces an EvaluationCriterion
    """

    def __init__(self, eval_func_entry: EvalFunctionEntry):
        self.eval_func_entry = eval_func_entry
        self.id = eval_func_entry.id
        self.name = eval_func_entry.name
        self.eval_func_arguments: Dict[str, Any] = {}

    def __repr__(self):
        return f"<EvalFunction: name={self.name}, id={self.id}>"

    @classmethod
    @abc.abstractmethod
    def expected_name(cls) -> str:
        """Name to look for in the EvalFunctionDefinitions"""

    def __call__(self, **kwargs) -> "EvalFunctionConfig":
        """Adding call to prepare for being able to pass parameters to function

        Notes:
            Technically now you could do something like eval_function > 0.5 but we want it
            to look like eval_function() > 0.5 to support eval_function(parameters) > 0.5
            in the future
        """
        self.eval_func_arguments.update(**kwargs)
        return self

    def __gt__(self, other) -> EvaluationCriterion:
        return self._op_to_test_metric(ThresholdComparison.GREATER_THAN, other)

    def __ge__(self, other) -> EvaluationCriterion:
        return self._op_to_test_metric(
            ThresholdComparison.GREATER_THAN_EQUAL_TO, other
        )

    def __lt__(self, other) -> EvaluationCriterion:
        return self._op_to_test_metric(ThresholdComparison.LESS_THAN, other)

    def __le__(self, other) -> EvaluationCriterion:
        return self._op_to_test_metric(
            ThresholdComparison.LESS_THAN_EQUAL_TO, other
        )

    def _op_to_test_metric(self, comparison: ThresholdComparison, value):
        return EvaluationCriterion(
            eval_function_id=self.eval_func_entry.id,
            threshold_comparison=comparison,
            threshold=value,
            eval_func_arguments=self.eval_func_arguments,
        )



import click
from rich.console import Console
from rich.table import Column, Table

from cli.client import init_client
from cli.helpers.nucleus_url import nucleus_url
from cli.helpers.web_helper import launch_web_or_invoke


@click.group("slices", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def slices(ctx, web):
    """Slices are named subsets of Datasets"""
    # TODO(gunnar): We don't have a natural landing for slices overview, until then we land on "/"
    launch_web_or_invoke("", ctx, web, list_slices)


@slices.command("list")
def list_slices():
    """List all available Slices"""
    client = init_client()
    console = Console()
    with console.status("Finding your Slices!", spinner="dots4"):
        table = Table(
            Column("id", overflow="fold", min_width=24),
            "name",
            "dataset_name",
            Column("url", overflow="fold"),
            title=":cake: Slices",
            title_justify="left",
        )
        datasets = client.datasets
        id_to_datasets = {d.id: d for d in datasets}
        all_slices = client.slices
        for s in all_slices:
            table.add_row(
                s.id,
                s.name,
                id_to_datasets[s.dataset_id].name,
                nucleus_url(f"{s.dataset_id}/{s.id}"),
            )

    console.print(table)
